// SPDX-License-Identifier: Unlicense
pragma solidity ^0.8.24;

import "./DAO.sol";
import "./IAcceptAvailability.sol";

struct AvailabilityStake {
    address worker;
    uint256 amount;
    uint endTime;
    bool assigned;
    bool reclaimed;
}

enum WorkStatus {
    Requested,
    EvidenceSubmitted,
    ApprovalSubmitted,
    Complete
}

struct WorkRequest {
    address customer;
    uint256 fee;
    WorkStatus status;
    uint stakeIndex;
    bool approval;
    uint poolIndex;
}

contract Work1 is IAcceptAvailability {
    DAO immutable dao;
    uint public immutable price;
    mapping(uint => AvailabilityStake) public stakes;
    uint public stakeCount;
    mapping(uint => WorkRequest) public requests;
    uint public requestCount;

    // TODO: Make parameters configurable
    uint constant POOL_DURATION = 1 days;

    event WorkAssigned(address worker, uint requestIndex);
    event WorkEvidenceSubmitted(uint requestIndex);

    constructor(DAO dao_, uint price_) {
        dao = dao_;
        price = price_;
    }

    /// Accept availability stakes as reputation token transfer
    function acceptAvailability(
        address sender,
        uint256 amount,
        uint duration
    ) external {
        AvailabilityStake storage stake = stakes[stakeCount++];
        stake.worker = sender;
        stake.amount = amount;
        stake.endTime = block.timestamp + duration;
    }

    function extendAvailability(uint stakeIndex, uint duration) external {
        AvailabilityStake storage stake = stakes[stakeIndex];
        require(
            msg.sender == stake.worker,
            "Worker can only extend their own availability stake"
        );
        stake.endTime = block.timestamp + duration;
    }

    function reclaimAvailability(uint stakeIndex) external {
        AvailabilityStake storage stake = stakes[stakeIndex];
        require(
            msg.sender == stake.worker,
            "Worker can only reclaim their own availability stake"
        );
        require(
            block.timestamp > stake.endTime,
            "Stake duration has not yet elapsed"
        );
        require(!stake.reclaimed, "Stake has already been reclaimed");
        stake.reclaimed = true;
        dao.transfer(msg.sender, stake.amount);
    }

    /// Select a worker randomly from among the available workers, weighted by amount staked
    function randomWeightedSelection() internal view returns (uint stakeIndex) {
        uint totalStakes;
        for (uint i = 0; i < stakeCount; i++) {
            if (stakes[i].assigned) continue;
            if (block.timestamp > stakes[i].endTime) continue;
            totalStakes += stakes[i].amount;
        }
        require(totalStakes > 0, "No available worker stakes");
        uint select = block.prevrandao % totalStakes;
        uint acc;
        for (uint i = 0; i < stakeCount; i++) {
            if (stakes[i].assigned) continue;
            if (block.timestamp > stakes[i].endTime) continue;
            acc += stakes[i].amount;
            if (acc > select) {
                stakeIndex = i;
                break;
            }
        }
    }

    /// Assign a random available worker
    function assignWork(uint requestIndex) internal returns (uint stakeIndex) {
        stakeIndex = randomWeightedSelection();
        AvailabilityStake storage stake = stakes[stakeIndex];
        stake.assigned = true;
        emit WorkAssigned(stake.worker, requestIndex);
    }

    /// Accept work request with fee
    function requestWork() external payable {
        require(msg.value >= price, "Insufficient fee");
        uint requestIndex = requestCount++;
        WorkRequest storage request = requests[requestIndex];
        request.customer = msg.sender;
        request.fee = msg.value;
        request.stakeIndex = assignWork(requestIndex);
    }

    /// Accept work evidence from worker
    function submitWorkEvidence(uint requestIndex) external {
        WorkRequest storage request = requests[requestIndex];
        require(
            request.status == WorkStatus.Requested,
            "Status must be Requested"
        );
        AvailabilityStake storage stake = stakes[request.stakeIndex];
        require(
            stake.worker == msg.sender,
            "Worker can only submit evidence for work they are assigned"
        );
        request.status = WorkStatus.EvidenceSubmitted;
        emit WorkEvidenceSubmitted(requestIndex);
    }

    /// Accept work approval/disapproval from customer
    function submitWorkApproval(uint requestIndex, bool approval) external {
        WorkRequest storage request = requests[requestIndex];
        require(
            request.status == WorkStatus.EvidenceSubmitted,
            "Status must be EvidenceSubmitted"
        );
        AvailabilityStake storage stake = stakes[request.stakeIndex];
        request.status = WorkStatus.ApprovalSubmitted;
        request.approval = approval;
        // Initiate validation pool
        request.poolIndex = dao.initiateValidationPool{value: request.fee}(
            stake.worker,
            POOL_DURATION
        );
    }
}
